@@ -7,15 +7,15 @@ use alloc::vec::Vec;
 use alloc::{format, vec};
 use core::fmt::{self, Debug, Formatter};
 
-use pki_types::{PrivateKeyDer, PrivatePkcs8KeyDer};
+use pki_types::{PrivateKeyDer, PrivatePkcs8KeyDer, SubjectPublicKeyInfoDer};
+use webpki::alg_id;
 
-use super::ring_like::io::der;
 use super::ring_like::rand::{SecureRandom, SystemRandom};
-use super::ring_like::signature::{self, EcdsaKeyPair, Ed25519KeyPair, RsaKeyPair};
+use super::ring_like::signature::{self, EcdsaKeyPair, Ed25519KeyPair, KeyPair, RsaKeyPair};
+use crate::crypto::signer::{public_key_to_spki, Signer, SigningKey};
 use crate::enums::{SignatureAlgorithm, SignatureScheme};
 use crate::error::Error;
-use crate::sign::{Signer, SigningKey};
-use crate::x509::{asn1_wrap, wrap_in_sequence};
+use crate::x509::{wrap_concat_in_sequence, wrap_in_octet_string};
 
 /// Parse `der` as any supported key encoding/type, returning
 /// the first which works.
@@ -123,6 +123,13 @@ impl SigningKey for RsaSigningKey {
             .map(|scheme| RsaSigner::new(Arc::clone(&self.key), *scheme))
     }
 
+    fn public_key(&self) -> Option<SubjectPublicKeyInfoDer<'_>> {
+        Some(public_key_to_spki(
+            &alg_id::RSA_ENCRYPTION,
+            self.key.public_key(),
+        ))
+    }
+
     fn algorithm(&self) -> SignatureAlgorithm {
         SignatureAlgorithm::RSA
     }
@@ -243,13 +250,10 @@ impl EcdsaSigningKey {
             _ => unreachable!(), // all callers are in this file
         };
 
-        let sec1_wrap = asn1_wrap(der::Tag::OctetString as u8, maybe_sec1_der);
+        let sec1_wrap = wrap_in_octet_string(maybe_sec1_der);
+        let pkcs8 = wrap_concat_in_sequence(pkcs8_prefix, &sec1_wrap);
 
-        let mut pkcs8_inner = Vec::with_capacity(pkcs8_prefix.len() + sec1_wrap.len());
-        pkcs8_inner.extend_from_slice(pkcs8_prefix);
-        pkcs8_inner.extend_from_slice(&sec1_wrap);
-
-        EcdsaKeyPair::from_pkcs8(sigalg, &wrap_in_sequence(&pkcs8_inner), rng).map_err(|_| ())
+        EcdsaKeyPair::from_pkcs8(sigalg, &pkcs8, rng).map_err(|_| ())
     }
 }
 
@@ -285,8 +289,18 @@ impl SigningKey for EcdsaSigningKey {
         }
     }
 
+    fn public_key(&self) -> Option<SubjectPublicKeyInfoDer<'_>> {
+        let id = match self.scheme {
+            SignatureScheme::ECDSA_NISTP256_SHA256 => alg_id::ECDSA_P256,
+            SignatureScheme::ECDSA_NISTP384_SHA384 => alg_id::ECDSA_P384,
+            _ => unreachable!(),
+        };
+
+        Some(public_key_to_spki(&id, self.key.public_key()))
+    }
+
     fn algorithm(&self) -> SignatureAlgorithm {
-        self.scheme.sign()
+        self.scheme.algorithm()
     }
 }
 
@@ -369,8 +383,12 @@ impl SigningKey for Ed25519SigningKey {
         }
     }
 
+    fn public_key(&self) -> Option<SubjectPublicKeyInfoDer<'_>> {
+        Some(public_key_to_spki(&alg_id::ED25519, self.key.public_key()))
+    }
+
     fn algorithm(&self) -> SignatureAlgorithm {
-        self.scheme.sign()
+        self.scheme.algorithm()
     }
 }
 
